// Copyright (c) 2016 Douglas Miller
import java.awt.*;
import javax.swing.*;
import java.io.*;
import java.net.*;
import java.lang.reflect.Field;
import java.util.Properties;
import java.util.Vector;
import java.util.Date;
import java.text.SimpleDateFormat;

public class CpnetSocketServer implements Runnable {
	private static CpnetSocketServer thus;
	private static final int vers = 1;

	public static void main(String[] args) {
		thus = new CpnetSocketServer(args);
	}

	Properties props = new Properties();
	int srvId = 0xff;
	String dir = null;
	ServerSocket sock;
	String host = null;
	int port = 0xff;
	String prefix;
	boolean magnet;

	public CpnetSocketServer(String[] args) {
		// TODO: need to redirect stderr...
		String portstr = null;
		String sidstr = null;
		String maxstr = "16";
		int max = 0;
		int sid = 0xff;
		char temp = 'P';
		String log = null;
		String rc = System.getenv("CPNET_CONFIG");
		if (rc == null) {
			File f = new File("./cpnetrc");
			if (f.exists()) {
				rc = f.getAbsolutePath();
			}
		}
		if (rc == null) {
			rc = System.getProperty("user.home") + "/.cpnetrc";
		}
		String s = props.getProperty("cpnetserver_log");
		if (s != null) {
			log = s;
		}
		// Command line args override config file...
		for (String arg : args) {
			if (arg.startsWith("conf=")) {
				rc = arg.substring(5);
			}
			if (arg.startsWith("log=")) {
				log = arg.substring(4);
			}
		}
		if (log != null) try {
			FileOutputStream fo = new FileOutputStream(log, false);
			PrintStream ps = new PrintStream(fo, true);
			System.setErr(ps);
		} catch (Exception ee) {}
		try {
			FileInputStream cfg = new FileInputStream(rc);
			props.load(cfg);
			cfg.close();
		} catch(Exception ee) {
			rc = null;
		}
		magnet = false;
		s = props.getProperty("cpnetserver_magnet");
		if (s != null) {
			magnet = true;
		}
		s = props.getProperty("cpnetserver_log_time");
		if (s != null) {
			PrintStream ps = new StampPrintStream(System.err);
			System.setErr(ps);
		}
		// At this point, "System.err" is ready to use...
		System.err.format("CpnetSocketServer v1.%d\n", vers);
		if (rc != null) {
			System.err.format("Using config in %s\n", rc);
		} else {
			System.err.format("No config file\n");
		}
		s = props.getProperty("cpnetserver_host");
		if (s != null) {
			host = s;
		}
		s = props.getProperty("cpnetserver_port");
		if (s != null) {
			portstr = s;
		}
		s = props.getProperty("cpnetserver_temp");
		if (s != null) {
			if (s.length() != 1 || !Character.isLetter(s.charAt(0))) {
				System.err.format("Invalid temp drive configured\n");
				System.exit(1);
			}
			temp = Character.toUpperCase(s.charAt(0));
		}
		s = props.getProperty("cpnetserver_sid");
		if (s != null) {
			sidstr = s;
		}
		s = props.getProperty("cpnetserver_max");
		if (s != null) {
			maxstr = s;
		}
		s = props.getProperty("cpnetserver_root_dir");
		if (s != null) {
			dir = s;
		}
		for (String arg : args) {
			if (arg.startsWith("host=")) {
				host = arg.substring(5);
			} else if (arg.startsWith("port=")) {
				portstr = arg.substring(5);
			} else if (arg.startsWith("temp=")) {
				String l = arg.substring(5);
				if (l.length() != 1 || !Character.isLetter(l.charAt(0))) {
					System.err.format("Invalid temp drive specified\n");
					System.exit(1);
				}
				temp = Character.toUpperCase(l.charAt(0));
			} else if (arg.startsWith("sid=")) {
				sidstr = arg.substring(4);
			} else if (arg.startsWith("max=")) {
				maxstr = arg.substring(4);
			} else if (arg.startsWith("dir=")) {
				dir = arg.substring(4);
			} else if (arg.equals("magnet")) {
				magnet = true;
			}
		}
		if (dir == null) {
			dir = System.getProperty("user.home") + "/HostFileBdos";
		}
		// TODO: mkdir? 
		if (portstr != null) {
			try {
				port = Integer.decode(portstr);
			} catch(Exception ee) {}
		}
		if (port <= 0 || host == null) {
			System.err.format("No valid host/port specified or configured\n");
			System.exit(1);
		}
		try {
			sid = Integer.valueOf(sidstr, 16);
		} catch(Exception ee) {}
		if (sid < 0 || sid >= 0xff) {
			System.err.format("No valid server ID specified or configured\n");
			System.exit(1);
		}
		try {
			max = Integer.valueOf(maxstr, 16);
		} catch(Exception ee) {}
		if (max <= 0) {
			max = 1;
		} else if (max > 16) {
			max = 16;
		}

		try {
			InetAddress ia;
			if (host.length() == 0 || host.equalsIgnoreCase("localhost")) {
				ia = InetAddress.getByName(null); // "localhost"
			} else {
				ia = InetAddress.getByName(host);
			}
			sock = new ServerSocket(port, 1, ia);
		} catch (Exception ee) {
			ee.printStackTrace(); // TODO: minimize
			System.exit(1);
		}

		// Each client gets a new instance of HostFileBdos,
		// but static data must be initialized first.
		srvId = sid;
		HostFileBdos.initCfg(temp, (byte)sid, max, dir);
		HostFileBdos.initLsts(props, "cpnetserver");
		// TODO: start up a GUI status/control frame?
		Thread t = new Thread(this);
		t.start();
	}

	private int dmp = 0;

	private void dumpMsg(byte[] msg, int len) {
		// always contains MAGNet header and CP/NET header?
		if (magnet) {
			dmp += ServerDispatch.getBC(msg);
			dmp &= 0x3ff;
			System.err.format("[%04x] %02x %04x %04x %04x:" +
					"%02x %02x %02x %02x %02x %02x ...\n",
				dmp,
				ServerDispatch.getCode(msg),
				ServerDispatch.getBC(msg),
				ServerDispatch.getDE(msg),
				ServerDispatch.getHL(msg),
				msg[NetworkServer.mcode] & 0xff,
				msg[NetworkServer.mdid] & 0xff,
				msg[NetworkServer.msid] & 0xff,
				msg[NetworkServer.mfunc] & 0xff,
				msg[NetworkServer.msize] & 0xff,
				msg[NetworkServer.mstart] & 0xff);
		} else {
			System.err.format("%02x %02x %02x %02x %02x:%02x %02x %02x %02x...\n",
				msg[NetworkServer.FMT] & 0xff,
				msg[NetworkServer.DID] & 0xff,
				msg[NetworkServer.SID] & 0xff,
				msg[NetworkServer.FNC] & 0xff,
				msg[NetworkServer.SIZ] & 0xff,
				msg[NetworkServer.DAT] & 0xff,
				msg[NetworkServer.DAT+1] & 0xff,
				msg[NetworkServer.DAT+2] & 0xff,
				msg[NetworkServer.DAT+3] & 0xff);
		}
	}

	class StampPrintStream extends PrintStream {
		PrintStream ps;
		SimpleDateFormat tagFmt = null;
		public StampPrintStream(PrintStream ps) {
			super(ps);
			this.ps = ps;
			tagFmt = new SimpleDateFormat("yyyy/MM/dd-HH:mm:ss.SSS");
		}
		@Override
		public PrintStream format(String fmt, Object ... parms) {
			fmt = tagFmt.format(new Date()) + " " + fmt;
			ps.format(fmt, parms);
			return this;
		}
	}

	class CpnetSocketServerConnection implements Runnable {
		Socket sock;
		InputStream in;
		OutputStream out;
		HostFileBdos hfb = null;
		NetworkBoot nwb;
		NetworkDump nwd;
		String remote;
		int cltId;
		byte[] buf;
		Properties props;

		public CpnetSocketServerConnection(Socket s) {
			// client MUST send us a string with client ID and our ID.
			// But, must receive that in thread.

			// TODO: pass in Properties, if we need them for boot/etc.
			props = new Properties();

			buf = new byte[256 + NetworkServer.mstart];
			int pt;
			sock = s;
			try {
				s.setKeepAlive(true);
				InetAddress ia = s.getInetAddress();
				remote = ia.getCanonicalHostName();
				pt = s.getLocalPort();
				in = sock.getInputStream();
				out = sock.getOutputStream();
			} catch (Exception ee) {
				ee.printStackTrace();
				return;
			}
			cltId = 0xff;
			System.err.format("Connection from %s (%d)\n", remote, pt);
			Thread t = new Thread(this);
			t.start();
		}

		private boolean newClient() {
			// first time, get client ID, etc.
			// assume this is a CP/Net message, so just extract
			// values.
			int cid;
			int sid;
			if (magnet) {
				cid = buf[NetworkServer.mDE] & 0xff;
				sid = buf[NetworkServer.mDEh] & 0xff; // should be us
			} else {
				cid = buf[NetworkServer.SID] & 0xff;
				sid = buf[NetworkServer.DID] & 0xff; // should be us
			}
			if (sid != srvId) {
				System.err.format("SID mismatch %02x != %02x\n",
					sid, srvId);
				return false;
			}
			if (cid == srvId || cid < 0 || cid >= 255) {
				System.err.format("CID invalid %02x\n",
					cid);
				return false;
			}
			cltId = cid;
			// TODO: maintain netTbl...
			System.err.format("Remote %s is %02x\n", remote, cltId);
			Vector<String> args = new Vector<String>();
			if (magnet) {
				hfb = new MagNetHFB(props, prefix, args, cid);
			} else {
				hfb = new HostFileBdos(props, prefix, args, cid);
			}
			return true;
		}

		private void shutdown() {
			try {
				sock.close();
			} catch (Exception eee) {}
			System.err.format("Disconnected from %s (%02x)\n", remote, cltId);
			if (hfb != null) {
				hfb.shutdown();
			}
		}

		private byte[] startBoot(byte[] buf) {
			nwb = NetworkBoot.doStart(props, buf);
			if (nwb == null) {
				return NetworkBoot.error(buf);
			}
			// TODO: some sort of timeout...
			// "nwb = null" when fires...
			return nwb.doNext(buf);
		}

		private byte[] startDump(byte[] buf) {
			nwd = NetworkDump.doStart(props, buf);
			if (nwd == null) {
				return NetworkDump.error(buf);
			}
			// TODO: some sort of timeout...
			// "nwd = null" when fires...
			return nwd.doNext(buf);
		}

		private byte[] processMsg(byte[] buf, int buflen) {
			byte[] rsp;
			int fmt = buf[NetworkServer.FMT] & 0xff;
			switch (fmt) {
			case 0x00: // CP/NET-80
				return hfb.sendMsg(buf, buflen);
			case 0xb0: // Boot
				if (nwb == null) {
					return startBoot(buf);
				}
				rsp = nwb.doNext(buf);
				if (nwb.done()) {
					nwb = null;
				}
				return rsp;
			case 0xd0: // Dump
				// TODO: how to prevent stuttering?
				if (nwd == null) {
					return startDump(buf);
				}
				rsp = nwd.doNext(buf);
				if (nwd.done()) {
					nwd = null;
				}
				return rsp;
			default:
				buf[NetworkServer.FMT] |= 1;
				byte src = buf[NetworkServer.SID];
				buf[NetworkServer.SID] = buf[NetworkServer.DID];
				buf[NetworkServer.DID] = src;
				buf[NetworkServer.SIZ] = 2 - 1;
				buf[NetworkServer.DAT] = (byte)0xff;
				buf[NetworkServer.DAT+1] = (byte)12;
				return buf;
			}
		}

		public void run() {
			int bufoff = 0;
			int buflen = 0;
			while (true) {
				int n = 0;
				try {
					n = in.read(buf, bufoff, buf.length - bufoff);
				} catch (Exception ee) {
					ee.printStackTrace();
					n = -1;
				}
				if (n < 0) {
					shutdown();
					return;
				}
				bufoff += n;
				if (buflen == 0) {
					if (!magnet) { // bare CP/NET
						if (bufoff < NetworkServer.mhdrlen) {
							continue;
						}
						buflen = buf[NetworkServer.SIZ] & 0xff;
						buflen += NetworkServer.mhdrlen + 1;
					} else {
						if (bufoff < NetworkServer.mpayload) {
							continue;
						}
						buflen = ServerDispatch.getBC(buf) +
							NetworkServer.mpayload;
					}
					if (hfb == null && !newClient()) {
						shutdown();
						return;
					}
				}
				if (buflen == 0 || bufoff < buflen) {
					continue;
				}
//if ((buf[NetworkServer.FMT] & 0xfe) == 0xd0)
				//dumpMsg(buf, buflen);
				// we have a message to process...
				byte[] rsp;
				int o = 0;
				if (!magnet) {
					rsp = processMsg(buf, buflen);
					n = (rsp[NetworkServer.SIZ] & 0xff) + 1;
					n += NetworkServer.DAT;
				} else {
					// 'hfb' is an MagNetHFB instance
					rsp = hfb.sendMsg(buf, buflen);
					int cd = ServerDispatch.getCode(rsp);
					n = NetworkServer.mpayload;
					if (cd == 0x01 || cd == 0x10) {
						n += ServerDispatch.getBC(rsp);
					} else if (cd == 0x30) {
						n += 65;
					}
				}
				try {
//if ((rsp[NetworkServer.FMT] & 0xfe) == 0xd0)
					//dumpMsg(rsp, n);
					out.write(rsp, o, n);
					out.flush();
				} catch (Exception ee) {
					ee.printStackTrace();
					shutdown();
					return;
				}
				// Reset for next loop
				bufoff = buflen = 0;
			}
		}
	}

	// Thread to listen on port and accept connections (clients).
	public void run() {
		Socket s;
		System.err.format("Server %02x Listening on %s port %d\n",
				srvId, host, port);
		while (true) {
			try {
				s = sock.accept();
				// must get client ID... but should not block
				// this thread... need new thread for HostFileBdos
				// anyway, so need some wrapper class.
				new CpnetSocketServerConnection(s);
				// if thread is running, it will keep object
				// from being released.
			} catch(IOException e) {
				break;
			}
		}
	}
}
