// Copyright (c) 2019 Douglas Miller <durgadas311@gmail.com>

import java.util.Arrays;
import java.util.Properties;
import java.util.Vector;
import java.util.Date;
import java.awt.event.*;
import java.io.*;
import java.text.SimpleDateFormat;

public class NetworkDump {
	// route all messages with this FMT to us:
	public static final int FMT = 0xd0;

	static final int RECLEN = 128;

	private static SimpleDateFormat timestamp =
		new java.text.SimpleDateFormat("yyyy-MM-dd-HH-mm-ss");

	private RandomAccessFile dump;

	private byte[] resp;

	public NetworkDump(byte[] msg, RandomAccessFile dmp) {
		dump = dmp;
		resp = new byte[128 + NetworkServer.DAT];
	}

	// If returns null, caller sends error(msg)
	public static NetworkDump doStart(Properties props, byte[] msg) {
		// Can't get here unless FMT == D0
//		if (msg[NetworkServer.FNC] != 1) {
//			return null;
//		}
		String dir = props.getProperty("netdump_dir");
		if (dir == null) {
			dir = System.getProperty("user.home") + "/NetDump";
		}
		// TODO: allow customization of name format...
		int nid = (msg[NetworkServer.SID] & 0xff);
		String t = timestamp.format(new Date());
		File file = new File(String.format("%s/cid_%02X-%s.dmp", dir, nid, t));
// TODO: fail-safe overwrite?
//		if (file.exists()) {
//			return null;
//		}
		try {
			RandomAccessFile f = new RandomAccessFile(file, "rw");
			System.err.format("Client %02x dumping to file \"%s\"\n",
				nid, file.getAbsolutePath());
			return new NetworkDump(msg, f);
		} catch (Exception ee) {
			return null;
		}
	}

	public static byte[] error(byte[] msg) {
		return new byte[]{ (byte)0xd1, msg[NetworkServer.SID],
				msg[NetworkServer.DID], 0, 0, 0 }; // NAK
	}

	public boolean done() { return (dump == null); }

	// Never returns null.
	public byte[] doNext(byte[] msg) {
		// Can't get here unless FMT == D0
		if (dump == null) {
			return error(msg);
		}
		int adr;
		int fnc = msg[NetworkServer.FNC] & 0xff;
		resp[NetworkServer.FNC] = (byte)0x01;	// ACK
		switch (fnc) {
		case 0x02:	// set "dma"...
			adr = (msg[NetworkServer.DAT] & 0xff) |
				((msg[NetworkServer.DAT+1] & 0xff) << 8);
			try {
				dump.seek(adr);
			} catch (Exception ee) {
				ee.printStackTrace();
				resp[NetworkServer.FNC] = (byte)0;	// NAK
				dump = null;
			}
			break;
		case 0x03:	// write record
			try {
				dump.write(msg, NetworkServer.DAT, RECLEN);
			} catch (Exception ee) {
				ee.printStackTrace();
				resp[NetworkServer.FNC] = (byte)0;	// NAK
				dump = null;
			}
			break;
		case 0x04:	// done
			try {
				dump.close();
			} catch (Exception ee) {
				ee.printStackTrace();
			}
			dump = null;
			break;
		default:	// error
			return error(msg);
		}
		// if reach here, send ack message...
		resp[NetworkServer.FMT] = (byte)0xd1;
		resp[NetworkServer.DID] = msg[NetworkServer.SID];
		resp[NetworkServer.SID] = msg[NetworkServer.DID];
		resp[NetworkServer.SIZ] = (byte)0;
		return resp;
	}
}
